#!/usr/bin/perl
#   Script Name: Fluffy the Pointless Forking Server
#        Author: Richard Terry
#       Website: http://radiac.net/perl
#       Version: 1.0.2
# Latest change: SIGHUP bug fixed, few small changes, better documentation

#############################
### Initialise the script ###

#####################
## Include modules ##
use Socket;
use POSIX;


######################
## Set up variables ##
&load_cmdlist; # Load commands
$fconfig = "fluffy.conf"; # Location of configuration file


#####################
## Predeclare subs ##
sub tellu;
sub out;
sub title;
sub AUTOLOAD { tellu " Cannot match command '$command'" }


#########################
## Read in config file ##
open(FILE,"$fconfig") || &error("Can't open the file '$fconfig'.<BR>Reason: $!\n");
@lines=<FILE>;
close(FILE);
foreach $line (@lines) {
	$line =~ s/\n|\r//;
	if (($line !~ /^\#/) && ($line =~ /=/)) {
		($var,$val) = split(/=/,$line,2);
		$config{$var} = $val;
	}
}


#########################
## Deal with arguments ##
foreach $arg (@ARGV) {
	if ($arg eq "-f" || $arg eq "--foreground") {
		$config{'daemon'} = "";
	} elsif ($arg eq "-b" || $arg eq "--background") {
		$config{'daemon'} = "yes";
	} elsif (($arg eq "-r" || $arg eq "--restart") || ($arg eq "-x" || $arg eq "--exit")) {
		if (not -e $config{'pidfile'}) {
			die "Cannot exit, no pidfile";
		} else {
			open(FILE,"$config{'pidfile'}") || &error("Can't open the file '$config{'pidfile'}'.<BR>Reason: $!\n");
			$pid = (<FILE>)[0];
			close(FILE);
			if ($arg eq "-x" || $arg eq "--exit") {
				kill TERM => $pid;
				print "SIGTERM sent.\n";
			} else {
				kill HUP => $pid;
				print "SIGHUP sent.\n";
			}
			exit;
		}
	}
}


########################
## Process management ##
# Ensure SIGHUP is unblocked
$sigset = POSIX::SigSet->new(SIGHUP);
sigprocmask(SIG_UNBLOCK, $sigset) or die "Can't unblock SIGHUP: $!\n";
fork and exit if ($config{'daemon'} eq "yes"); # If we want to run in the background, now's the time... but we don't want it to fork on a reload
$PROGRAM_NAME = "Fluffy Server";
$SIG{CHLD} = 'IGNORE'; # We're not interested in children's corpses; this line stops zombie processes
$SIG{HUP} = \&sighup; # We want to be able to reload
$SIG{TERM} = \&sigterm; # We want to control the termination
$parent = $$;


######################################
## Open socket ready for connection ##
select(CLIENT); # Select CLIENT channel
$| = 1; # Force a flush after every write to the channel
select(STDOUT); # Select STDOUT channel again

socket(SERVER, PF_INET, SOCK_STREAM, getprotobyname('tcp')); # Make the socket
setsockopt(SERVER, SOL_SOCKET, SO_REUSEADDR, 1); # Get over the delay for releasing it
$sockaddr = sockaddr_in($config{'port'}, INADDR_ANY); # Socket address
bind(SERVER, $sockaddr) or die "Couldn't bind to port $config{'port'}: $!\n";
listen(SERVER, SOMAXCONN) or die "Couldn't listen on port $config{'port'}: $!\n"; # Start incoming queue

# We are running as server...
if ($config{'pidfile'} ne "") {
	open (PID,">$config{'pidfile'}") || die "Couldn't open log file - $!\n";
	print PID $parent;
	close PID;
}

# If we're going to be logging, start logging now
if ($config{'outpipe'} ne "") {
	open (FILE,">$config{'outpipe'}") || die "Couldn't open log file - $!\n";
	select(FILE); # Select CLIENT channel
	$| = 1; # Force a flush after every write to the channel
}

# Send output to say we're ready...
out "^gInitialised^n as^y $$";
out "^gSocket open^n on port $config{'port'}";


############################
### Main forking section ###
# Accept and process connections
while ($client_address = accept(CLIENT, SERVER)) {
	my ($in);
	if (!defined($child_pid = fork())) {
		out "^RTerminating^n - could not fork";
		exit;
	} elsif ($child_pid) {
    	# I'm the parent
		out "^gForking^y $child_pid...";
	} else {
	    # I'm the child
		$PROGRAM_NAME = "Fluffy Child";
		($port, $packed_ip) = sockaddr_in($client_address);
		$dotted_quad = inet_ntoa($packed_ip);
		out "^gForked^n - connection from ^p$dotted_quad $port^n";
		title "Fluffy the Pointless Forking Server";
		tellu "^wHello and welcome to ^yFluffy^w the ^pPointless ^cForking ^gServer^w!";
		tellu "Type 'commands' to list available commands, and enjoy your stay...\n^L";
		while ($in=<CLIENT>) {
			# A ' at the start of the line will cause an error!
			while ($in =~ s/^'//) {}
			$in =~ s/\n|\r//g;
			($command, @vars) = split(/ /, $in);
			my $tc = "command_$command";
			&$tc(@vars);
		}
		out "^rTerminating^n - Client disconnected";
		close CLIENT;
		exit;
	}
}
out "^rTerminating^n - No more connections";
close SERVER;
close FILE;
exit;

###################
### Subroutines ###

##############
## Commands ##

sub load_cmdlist {
	# Hash of commands and descriptions for use with command_commands and command_describe
	%cmdlist = (
	"commands", "List all commands",
	"colours", "Shows the different colours available",
	"describe", "Describes a command",
	"quit", "Disconnects you from the program"
	);
}

sub command_commands {
	title "Commands";
	foreach $key (sort keys %cmdlist) {
		tellu "$key";
	}
	tellu "^L";
}

sub command_colours {
	title "Available Colours";
	my @colours = qw(dark red green yellow blue purple cyan white normal);
	foreach $colour (@colours) {
		$colour =~ /^(.)/;
		tellu "$colour (^^$1): ^$1hello";
	}
	tellu "And don't forget, ^^L will give you a line.\n^L";
}

sub command_describe {
	my ($command) = shift(@vars);
	unless (exists($cmdlist{$command})) {
		tellu " Cannot match command '$command'";
		return;
	}
	title $command;
	tellu "$cmdlist{$command}\n^L";
}

sub command_egg {
	tellu "Woo! You found the easter egg!";
}

sub command_quit {
	title "Goodbye!";
	tellu "Bye... come again soon!\n^L\n";
	out "^rTerminating^n - Client quit";
	close CLIENT;
	exit;
}

#######

sub out {
	my $out = "^c$$^n> $_[0]";
	if ($config{'logcolour'} eq "yes") {
		$out = &trans($out);
	} else {
		$out =~ s/\^[drgybpcwn]//g;
		$out = &trans($out);
	}
	print "$out\n";
}

sub tellu {
	my $out = &trans($_[0]);
	$out =~ s/\n/\r\n/g;
	print CLIENT "$out\r\n";
}

sub prompt {
	my $out = &trans($_[0]);
	$out =~ s/\n/\r\n/g;
	print CLIENT "$out";
	$answer = <CLIENT>;
	$answer =~ s/\n|\r//g;
	return $answer;
}

sub trans {
	my $out = $_[0];
	
	# Dynatext
	$out =~ s/(^|[^\^])\^L/$1^y$config{'line'}\e[0m/g; # Line
	
	# Colours
	my $col = &col($out);
	$out = $col . "\e[0m" if ($col ne $out);
	
	$out =~ s/\^\^/\^/g;
	
	return $out;
}

sub col {
	# Codes from http://www.linuxgazette.com/issue01to08/articles.html
	my $c = $_[0];
	$c =~ s/(^|[^\^])\^d/$1\e[1;30m/g; # black
	$c =~ s/(^|[^\^])\^r/$1\e[1;31m/g; # red
	$c =~ s/(^|[^\^])\^g/$1\e[1;32m/g; # green
	$c =~ s/(^|[^\^])\^y/$1\e[1;33m/g; # yellow
	$c =~ s/(^|[^\^])\^b/$1\e[1;34m/g; # blue
	$c =~ s/(^|[^\^])\^p/$1\e[1;35m/g; # purple
	$c =~ s/(^|[^\^])\^c/$1\e[1;36m/g; # cyan
	$c =~ s/(^|[^\^])\^w/$1\e[1;37m/g; # white
	$c =~ s/(^|[^\^])\^n/$1\e[0m/g; # normal
	return $c;
}

sub title {
	my $in = $_[0];
	return $in if (length($in) > 74);
	$in = " $in ";
	my $out = $config{'line'};
	substr($out, (length($out)/2)-(length($in)/2), length($in)) = $in;
	$out =~ s/ ([\=\-\*]+)$/ \^y$1/ if ($in =~ /\^/);
	tellu "\n\r^y$out";
}

sub sighup {
	$SIG{HUP} = \&sighup; # We want to be able to reload
	
	if ($parent == $$) {
		# Tell the log what's happening...
		out "^pRestarting^n - ^pSIGHUP^n recieved";
		
		# Close the port
		close SERVER;
		
		# Find out the file address and run ourselves again
		$fself .= $ENV{'_'};
		$fself =~ s/^\.//;
		exec "$ENV{'PWD'}$fself", "-f";
		exit;
		
	} else {
		# We're a child, we're not doing anything, because we don't have to
		out "Childishly stubborn - ^pSIGHUP^n recieved";
	}
}

sub sigterm {
	out "^rTerminating^n - ^rSIGTERM^n recieved";
	if ($parent == $$) {
		unlink($config{'pidfile'}) unless ($config{'pidfile'} eq "");
	} else {
		tellu " ^rTerminating^n - ^rSIGTERM^N recieved\n";
	}
	exit;
}
